@@ -5,6 +5,7 @@ from .lattice import Lattice
 from .physics import *
 from .result import *
 from .calc import *
+from .heterostructure_opt import *
 
 
 class Heterostructure:
@@ -32,7 +33,7 @@ class Heterostructure:
     #
     # MN – basis change matrix from N to M
     #      (the order of the letters makes it easy to combine these:
-    #       MM' @ M'N == MN)
+    #       MW @ WN == MN for any W)
     # v_M – vector v (unit vector of basis V) in basis M
     # v_Ms – an array of vectors v in basis M
     # stg_lay – list of "stg" for each of the layers
@@ -51,6 +52,8 @@ class Heterostructure:
     #          vectors
     # X, x – cartesian basis (unit vectors: (1 angstrom, 0),
     #                                    (0, 1 angstrom))
+    # Xt, xt – Transformation corresponding to Br - Btr transformation, but in
+    # cartesian basis
     #
     # Note that the vector space of all the mentioned objects is R^2
 
@@ -246,7 +249,7 @@ class Heterostructure:
         Parameters
         ----------
         XBr : Matrix 2x2
-        XBr : Matrix 2x2
+        XXt : Matrix 2x2
 
         Returns
         -------
@@ -323,6 +326,45 @@ class Heterostructure:
         ADt = np.array(M)
         return self.__calc_aux(ADt, thetas)
 
+    def __calc_strain_tensor_from_ADt(self, ADt: Matrix2x2, XBr: Matrix2x2) -> Matrix2x2:
+        """
+        Calculate strain tensor.
+
+        See docs of `Heterostructure.calc` for definition of strain tensor.
+
+        Parameters
+        ----------
+        ADt : Matrix 2x2
+        XBr : Matrix 2x2
+
+        Returns
+        -------
+        Matrix 2x2
+        """
+        BrDt, BtrBr, XXt = self.__get_basis_change_matrices(ADt, XBr)
+        return Heterostructure.__calc_strain_tensor(XBr, XXt)
+
+    def __get_basis_change_matrices(self, ADt, XBr):
+        """
+        Helper function to retrieve BrDt, BtrBr, XXt matrices
+
+        Parameters
+        ----------
+        ADt : Matrix 2x2
+        XBr : Matrix 2x2
+
+        Returns
+        -------
+        BrDt : Matrix 2x2
+        BtrDt : Matrix 2x2
+        XXt : Matrix 2x2
+        """
+        XA = self.__substrate.basis_change_matrix()[0:2, 0:2]
+        BrDt = inv(XBr) @ XA @ ADt
+        BtrBr = Heterostructure.__get_BtrBr(BrDt)
+        XXt = XBr @ inv(BtrBr) @ inv(XBr)
+        return BrDt, BtrBr, XXt
+
     def __calc_aux(self,
                    ADt: InMatrix2x2,
                    thetas: List[Angle]) -> Result:
@@ -343,15 +385,20 @@ class Heterostructure:
         Result
             see `Result` documentation for details
         """
-
         XA = self.__substrate.basis_change_matrix()[0:2, 0:2]
         XBrs = [rotate(lay_desc[0].basis_change_matrix()[0:2, 0:2], theta)
                 for theta, lay_desc in zip(thetas, self.__layers)]
-        BrDts = [inv(XBr) @ XA @ ADt for XBr in XBrs]
-        BtrBrs = [Heterostructure.__get_BtrBr(BrDt) for BrDt in BrDts]
-        XXts = [XBr @ inv(BtrBr) @ inv(XBr) for XBr, BtrBr in zip(XBrs, BtrBrs)]
-        strain_tensors = [Heterostructure.__calc_strain_tensor(XBr, XXt)
-                          for XBr, XXt in zip(XBrs, XXts)]
+        strain_tensors = [self.__calc_strain_tensor_from_ADt(ADt, XBr)
+                          for XBr in XBrs]
+        BrDts = []
+        BtrBrs = []
+        XXts = []
+        for XBr in XBrs:
+            BrDt, BtrBr, XXt = self.__get_basis_change_matrices(ADt, XBr)
+            BrDts.append(BrDt)
+            BtrBrs.append(BtrBr)
+            XXts.append(XXt)
+
         ABtrs = [inv(XA) @ XBr @ inv(BtrBr) for XBr, BtrBr in zip(XBrs, BtrBrs)]
 
         # also add the alternative strain tensor definition to the Result
@@ -518,6 +565,9 @@ class Heterostructure:
 
     @staticmethod
     def __get_BtrBr(BrDt):
+        """
+        BrDt -> BtrBr
+        """
         # see why round in implementation of `Heterostructue.__get_d_xs`
         BtrDt = np.round(BrDt)
         return BtrDt @ inv(BrDt)
@@ -526,7 +576,8 @@ class Heterostructure:
 
     def opt(self,
             max_el: int = 6,
-            thetas: Optional[List[Optional[List[float]]]] = None
+            thetas: Optional[List[Optional[List[float]]]] = None,
+            algorithm: str = "moire"
             ) -> Result:
         """
         Minimises strain, and calculates its value.
@@ -550,17 +601,34 @@ class Heterostructure:
             of floats, without the need for nesting.
             All angles are in radians.
 
+        algorithm : str, optional
+            Default: "moire"
+            Accepted values: "moire", "brute"
+
         Returns
         -------
         Result
             Result object containing the results of the optimisation.
             For more information see documentation of `Result`
         """
+
+        # TODO: the whole routine is pretty complex;
+        #   maybe separate opt numerics into a class "MoireSolver" or stg.
+        #   so we don't have to deal with passing XA, XB etc. all the time,
+        #   and the mess in Heterostructure
+
         # Prepare ranges of theta values
+        def is_iterable(x):
+            try:
+                iter(x)
+                return True
+            except Exception:
+                return False
+
         if thetas is not None:
-            if isinstance(thetas[0], list):
+            if is_iterable(thetas[0]):
                 thetas_in = [arg if arg is not None else np.arange(*lay_desc[1])
-                            for arg, lay_desc in zip(thetas, self.__layers)]
+                             for arg, lay_desc in zip(thetas, self.__layers)]
             elif len(self.__layers) == 1:
                 thetas_in = [thetas]
             else:
@@ -570,194 +638,19 @@ class Heterostructure:
 
         # Using (1, 1) norm since we are usually interested in minimising
         # |strain_ij| over i, j (where 'strain' is strain tensor)
-        thetas, ADt = self.__opt_aux((1, 1), max_el, thetas_in)
+
+        config = OptSolverConfig()
+        config.max_el = max_el
+        config.ord = (1, 1)
+
+        XA, XBs = self.__get_lattice_matrices()
+        if algorithm == "moire":
+            thetas, ADt = MoireFinder(XA, XBs, thetas_in, config).solve()
+        else:  # "brute"
+            thetas, ADt = StrainOptimisator(XA, XBs, thetas_in, config).solve()
 
         return self.calc(ADt, thetas)
 
-    def __opt_aux(self,
-                  ord: Tuple[int, int],
-                  max_el: int,
-                  thetas_in: List[List[Angle]]) \
-            -> Tuple[List[Angle], Matrix2x2]:
-        """
-        This routine calculates optimal supercell lattice vectors, layers'
-        rotation angles and their strain tensors. Here, optimal means
-        values that result in L_{1, 1} strain tensor norm to be smallest
-        :math:`L_{11}(\epsilon) = \max_{ij} |\epsilon_{ij}\` [1]
-
-        Parameters
-        ----------
-        ord : (int, int)
-            (p, q) for calculating L_{p, q} norm of the strain tensor
-        max_el : int
-            Maximum absolute value of ADt matrix element
-        thetas_in : List[List[float]]
-            Must have length equal to the number of layers in the heterostructure.
-            Elements are lists or arrays containing possible values of thata for
-            a given layer.
-
-        Returns
-        -------
-        List[float]
-            List of optimal theta values corresponding to the Heterostructure
-            layers. Length is equal to the number of layers.
-        Matrix2x2
-            Best ADt matrix found
-        """
-
-        # prepare layer matrices
-        XA, XB_lay = self.__get_lattice_matrices()
-
-        dt_As = self.__get_dt_As(max_el)
-
-        # Dummy start value for classic O(n) find min algorithm, we want to find
-        # values for which sum of `quality_fun`(strain tensor) is minimal
-        res: Tuple[List[Angle], float, Matrix2x2, List[Matrix2x2]] = \
-            (None, np.inf, None, None)
-
-        # embarrasingly parallel, but Python GIL makes this irrelevant
-        for theta_lay in itertools.product(*thetas_in):
-            strain_tensor_lay = [Heterostructure.__get_strain_tensor_opt(
-                theta, XA, XB, dt_As
-            ) for theta, XB in zip(theta_lay, XB_lay)]
-
-            # qty – array which contains norms of the strain tensors
-            qty = sum([matnorm(st, ord[0], ord[1]) for st in strain_tensor_lay])
-
-            # if qty is NaN it means that we somehow ended up with linear
-            # dependence; in the limit strain would go to infinity
-            qty[np.isnan(qty)] = np.inf
-
-            argmin_indices = np.unravel_index(qty.argmin(), qty.shape)
-            min_qty = qty[argmin_indices]
-            min_st_lay = [st[argmin_indices] for st in strain_tensor_lay]
-            ADt = np.stack((dt_As[argmin_indices[0], argmin_indices[2]],
-                            dt_As[argmin_indices[1], argmin_indices[3]]))
-            XDt = XA @ ADt
-
-            # let's check if the best values for this combination of theta vals
-            # (theta_lay) are better than those we already have
-            res = Heterostructure.__update_opt_res(res, theta_lay, min_qty, XDt, min_st_lay)
-
-        ADt = inv(XA) @ res[2]
-        thetas = res[0]
-        return thetas, ADt
-
-    @staticmethod
-    def __get_strain_tensor_opt(theta: float,
-                                XA: Matrix2x2,
-                                XB: Matrix2x2,
-                                dt_As: np.ndarray) -> np.ndarray:
-        """
-        Calculates strain tensor for `opt_aux`.
-        Just lots of linear algebra, really.
-
-        Parameters
-        ----------
-        theta : float
-        XA : Matrix 2x2
-        XB : Matrix 2x2
-        dt_As : np.ndarray, shape (span, span, 2)
-
-        Returns
-        -------
-        np.ndarray with shape (span, span, span, span, 2, 2)
-
-        Notes
-        -----
-        Definiton of strain tensor here is the same as in documentation
-        for `calc`
-        """
-
-        dt_xs = matvecmul(XA, dt_As)
-        AX = inv(XA)
-        d_xs = Heterostructure.__get_d_xs(AX, XB, dt_As, theta)
-        XXt = Heterostructure.__get_XXt(d_xs, dt_xs)
-        XBr = rotate(XB, theta)
-        return Heterostructure.__calc_strain_tensor(XBr, XXt)
-
-    @staticmethod
-    def __get_dt_As(max_el: int) -> np.ndarray:
-        """
-        Prepares dt_As for __opt_aux calculation.
-
-        Parameters
-        ----------
-        max_el : int
-            Maximum absolute value of ADt matrix element
-
-        Returns
-        -------
-        dt_As : np.ndarray, shape (span, span, 2)
-        """
-        # Let's imagine all possible superlattice vectors (dt vectors). At the
-        # very least, they need to be able to recreate the substrate lattice.
-        # We assume substrate lattice to be constant.
-        # Thus every grid point in substrate basis (A) is a valid dt vector
-        span_range = np.arange(0, 2 * max_el + 1)
-        span_range[(max_el + 1):] -= 2 * max_el + 1
-        dt_As = np.transpose(np.meshgrid(span_range, span_range))
-
-        return dt_As
-
-    @staticmethod
-    def __update_opt_res(res: Tuple[List[Angle], float, Matrix2x2, List[Matrix2x2]],
-                         thetas: Tuple[Angle, ...],
-                         min_qty: float,
-                         XDt: np.ndarray,
-                         min_st: List[Matrix2x2]
-                         ) -> Tuple[List[Angle], float, Matrix2x2, List[Matrix2x2]]:
-        """
-        Checks if newly calculated result is better than the previous one
-        (this usually means that the strain measure `qty` is smaller),
-        and if so updates it accordingly
-
-        Parameters
-        ----------
-        res : List[float], float, Matrix2x2, List[Matrix2x2]
-        thetas : Collection[float]
-        min_qty : float
-        XDt : Matrix2x2
-        min_st : List[Matrix2x2]
-
-        Returns
-        -------
-        List[float]
-        float
-        Matrix2x2
-        List[Matrix2x2]
-        """
-
-        # 1. Check for smaller supercell quality function
-
-        if min_qty - ABS_EPSILON > res[1]:
-            return res
-
-        thetas = list(thetas)
-        new_res = (thetas, min_qty, XDt, min_st)
-
-        # Here it is most efficient to check whether we aren't at the beginning
-        # (res[0] is None in that case)
-        if (min_qty + ABS_EPSILON < res[1]) or (res[0] is None):
-            return new_res
-
-        # 2. If qties are almost equal, choose smaller elementary cell
-
-        old_size = np.abs(np.linalg.det(res[2]))
-        new_size = np.abs(np.linalg.det(XDt))
-
-        if new_size < old_size - ABS_EPSILON:
-            return new_res
-        elif old_size < new_size - ABS_EPSILON:
-            return res
-
-        # 3. If even the sizes are equal, return the cell more 'square-y'
-        # here it will mean the cell with smaller |max_lattice_vector_element|
-
-        if np.max(np.abs(XDt)) < np.max(np.abs(res[2])):
-            return new_res
-        return res
-
     def __get_lattice_matrices(self):
         """
         Calculates basis change vectors between Cartesian basis and
@@ -774,76 +667,6 @@ class Heterostructure:
                for lay_desc in self.__layers]
         return XA, XBs
 
-    @staticmethod
-    def __get_d_xs(
-            AX: np.ndarray,
-            XB: np.ndarray,
-            dt_As: np.ndarray,
-            theta: Angle) -> np.ndarray:
-        """
-        Returns an array of `d` vectors in Cartesian basis
-
-        Parameters
-        ----------
-        AX : Matrix2x2
-        XB : Matrix 2x2
-        dt_As : np.ndarray, shape (..., 2)
-        theta : float
-
-        Returns
-        -------
-        np.ndarray, shape (..., 2)
-        """
-        XBr = rotate(XB, theta)
-        BrA = inv(AX @ XBr)
-
-        dt_Brs = matvecmul(BrA, dt_As)
-
-        # Here we use the fact that the supercell must "stretch" lattice vectors
-        # of constituent layers so that they superlattice vectors are linear
-        # _integer_ combinations of any one layers' lattice vectors.
-        dt_Btrs = np.round(dt_Brs)
-
-        d_Brs = dt_Btrs
-        return matvecmul(XBr, d_Brs)
-
-    @staticmethod
-    def __get_XXt(d_xs: np.ndarray, dt_xs: np.ndarray) -> np.ndarray:
-        """
-        Calculates basis change matrices Xt -> X
-
-        Parameters
-        ----------
-        dt_xs : np.ndarray with shape (span, span, 2)
-            array of dt(x, y) vectors in X basis
-        d_xs : np.ndarray with shape (span, span, 2)
-            array of d(x, y) vectors in X basis
-
-        Returns
-        -------
-        np.ndarray with shape (span, span, span, span, 2, 2)
-            array of XXt matrices, where element XXt[x1, y1, x2, y2]
-            corresponds to d(x1, y1) and d(x2, y2) vectors.
-            If vectors d(x1, y1) and d(x2, y2) make a singular matrix,
-            corresponding values are changed to np.inf
-        """
-        span = d_xs.shape[0]
-
-        XDt = np.empty((span, span, span, span, 2, 2))
-        XDt[..., 0] = dt_xs[np.newaxis, np.newaxis, ...]
-        XDt[..., 1] = dt_xs[..., np.newaxis, np.newaxis, :]
-
-        XD = np.empty((span, span, span, span, 2, 2))
-        XD[..., 0] = d_xs[np.newaxis, np.newaxis, ...]
-        XD[..., 1] = d_xs[..., np.newaxis, np.newaxis, :]
-
-        # what we want is basically XDt @ DX, which means we need to invert XD
-        # (XXt = XD DDt DX = XDt DX)
-
-        return XDt @ inv(XD)
-
-        # TODO: suppress divide-by-zero warning
-
 
 def heterostructure() -> Heterostructure:
     """
